"""
Training of CGAN
"""

import torch
import torch.nn as nn
import torch.optim as optim
from torch.utils.data import DataLoader, TensorDataset
from sklearn.svm import SVC
from sklearn.neural_network import MLPClassifier
from sklearn.ensemble import RandomForestClassifier
from sklearn.metrics import accuracy_score
from sklearn.metrics import f1_score
import matplotlib.pyplot as plt
import numpy as np
from utils import gradient_penalty, save_checkpoint, load_checkpoint, logToFile
from model import DiscriminatorVanillia, Generator, weights_init
from dataset import load_subject_transfer_data, load_session_transfer_data

# Hyperparameters etc.
device = "cuda" if torch.cuda.is_available() else "cpu"
LEARNING_RATE = 2e-4
WEIGHT_DECAY = 1e-6
BATCH_SIZE = 20
NUMBER_OF_SUBJECTS = 16
NUMBER_OF_SESSIONS = 2
EEG_CHANNELS = 32
EEG_BANDS = 4
PERIPHERAL_CHANNELS = 8
PERIPHERAL_BANDS = 8
Z_DIM = 100
GEN_INPUT = Z_DIM + PERIPHERAL_CHANNELS*PERIPHERAL_BANDS + 1
GEN_OUTPUT = EEG_CHANNELS*EEG_BANDS
GEN_DIM = (128, 128)
DISC_DIM = (64, 64)
DISC_INPUT = GEN_OUTPUT + 1
NUM_CLASSES = 2
NUM_EPOCHS = 300
CRITIC_ITERATIONS = 5
LAMBDA_GP = 10

def train(transfer_method,idx,session):
    if transfer_method == "session-based":
        EEG_data, EEG_label, EYE_data, EYE_label, EEG_gen_test, EEG_gen_label, EYE_gen_test, EYE_gen_label, EEG_svc_test, EEG_svc_label, EYE_svc_test, EYE_svc_label = load_session_transfer_data(idx,session)
    else:
        EEG_data, EEG_label, EYE_data, EYE_label, EEG_gen_test, EEG_gen_label, EYE_gen_test, EYE_gen_label, EEG_svc_test, EEG_svc_label, EYE_svc_test, EYE_svc_label = load_subject_transfer_data(idx,session)
    
    print("EEG:",EEG_data.shape,"EYE:",EYE_data.shape," label:",EEG_label.shape)
    train_dataset = TensorDataset(torch.FloatTensor(EEG_data), torch.FloatTensor(EYE_data), torch.LongTensor(EEG_label))
    dataloader = DataLoader(train_dataset, batch_size = BATCH_SIZE, shuffle = True)

    gen_test_eye = torch.FloatTensor(EYE_gen_test)
    gen_test_label = torch.LongTensor(EYE_gen_label)
    gen_test_input = torch.cat([gen_test_eye, gen_test_label.reshape(-1,1)], dim=1).to(device)

    # initialize gen and disc, note: discriminator should be called critic,
    # according to WGAN paper (since it no longer outputs between [0, 1])
    gen = Generator(GEN_INPUT, GEN_DIM, GEN_OUTPUT).to(device)
    critic = DiscriminatorVanillia(DISC_INPUT, DISC_DIM).to(device)
    weights_init(gen)
    weights_init(critic)

    # initializate optimizer
    opt_gen = optim.Adam(gen.parameters(), lr=LEARNING_RATE, betas=(0.5, 0.9), weight_decay=WEIGHT_DECAY)
    opt_critic = optim.Adam(critic.parameters(), lr=LEARNING_RATE, betas=(0.5, 0.9), weight_decay=WEIGHT_DECAY)

    criterion = nn.BCELoss() # binary cross entropy loss

    step = 0
    gen_losses = []
    disc_losses = []
    gen_loss_avg = []
    disc_loss_avg = []

    if session == 1 or transfer_method == "session-based":

        gen.train()
        critic.train()

        for epoch in range(NUM_EPOCHS):
            for batch_idx, (real_eeg, real_eye, labels) in enumerate(dataloader):
                real_eeg = real_eeg.to(device)
                real_eye = real_eye.to(device)
                cur_batch_size = real_eeg.shape[0]
                labels = labels.to(device)
                # generate fake data
                noise = torch.randn(cur_batch_size, Z_DIM).to(device)
                gen_input = torch.cat([noise, real_eye, labels.reshape(-1,1)], dim=1)
                disc_input = torch.cat([real_eeg, labels.reshape(-1,1)], dim=1)
                fake = gen(gen_input)
                ### Train Discriminator: max log(D(x)) + log(1 - D(G(z)))
                fake_input = torch.cat([fake, labels.reshape(-1,1)], dim=1)
                critic_real = critic(disc_input).reshape(-1)
                critic_fake = critic(fake_input).reshape(-1)

                loss_disc_real = criterion(critic_real, torch.ones_like(critic_real))
                loss_disc_fake = criterion(critic_fake, torch.zeros_like(critic_fake))
                loss_critic = (loss_disc_real + loss_disc_fake) / 2

                critic.zero_grad()
                loss_critic.backward(retain_graph=True)
                opt_critic.step()

                ### Train Generator: min log(1 - D(G(z))) <-> max log(D(G(z))
                gen_fake = critic(fake_input).reshape(-1)
                loss_gen = criterion(gen_fake, torch.ones_like(gen_fake))
                gen.zero_grad()
                loss_gen.backward()
                opt_gen.step()

                gen_loss_avg.append(loss_gen.item())
                disc_loss_avg.append(loss_critic.item())

            # Print losses occasionally and print to tensorboard
            if epoch % 10 == 0 and epoch > 0:
                print(
                    f"Epoch [{epoch}/{NUM_EPOCHS}] \
                        Loss D: {loss_critic:.4f}, loss G: {loss_gen:.4f}"
                )

                with torch.no_grad():
                    eval_batch_size = gen_test_input.shape[0]
                    noise = torch.randn(eval_batch_size, Z_DIM).to(device)
                    gen_eval_input = torch.cat([noise, gen_test_input], dim=1)
                    eeg_fake = gen(gen_eval_input)
                    eeg_fake = eeg_fake.cpu().detach().numpy()
                    X_train = np.concatenate([eeg_fake, EYE_gen_test], axis=1)
                    # train and test
                    model = SVC(C=2, kernel = 'rbf', degree=3)
                    model.fit(X_train, EYE_gen_label)
                    X_test = np.concatenate([EEG_svc_test, EYE_svc_test], axis=1)
                    yhat = model.predict(X_test)
                    print("Fake Accuracy:", accuracy_score(y_true=EYE_svc_label, y_pred = yhat))
                    
                    save_checkpoint(gen, opt_gen, filename="cwgan_gen.pth.tar")
                    save_checkpoint(critic, opt_critic, filename="cwgan_critic.pth.tar")

                step += 1
            gen_losses += [np.mean(gen_loss_avg)]
            disc_losses += [np.mean(disc_loss_avg)]

        save_checkpoint(gen, opt_gen, filename="cwgan_gen.pth.tar")
        save_checkpoint(critic, opt_critic, filename="cwgan_critic.pth.tar")

        #plt.plot(gen_losses, label = "Generator Loss")
        #plt.plot(disc_losses, label = "Discriminator Loss")
        #plt.legend()
        #plt.show()

    load_checkpoint("cwgan_gen.pth.tar", gen, opt_gen, LEARNING_RATE)
    load_checkpoint("cwgan_critic.pth.tar", critic, opt_critic, LEARNING_RATE)

    with torch.no_grad():
        gen.eval()
        eval_batch_size = gen_test_input.shape[0]
        noise = torch.randn(eval_batch_size, Z_DIM).to(device)
        gen_eval_input = torch.cat([noise, gen_test_input], dim=1)
        eeg_fake = gen(gen_eval_input)
        eeg_fake = eeg_fake.cpu().detach().numpy()
        X_train = np.concatenate([eeg_fake, EYE_gen_test], axis=1)
        # train and test
        model = SVC(C=2, kernel = 'rbf', degree=3)
        model.fit(X_train, EYE_gen_label)
        X_test = np.concatenate([EEG_svc_test, EYE_svc_test], axis=1)
        yhat = model.predict(X_test)
        acc_fake_svc = accuracy_score(y_true=EYE_svc_label, y_pred = yhat)
        print("Fake Accuracy SVC:", acc_fake_svc)
        model = MLPClassifier( activation='relu',max_iter=1000, random_state=1, alpha = 0.001, batch_size= 10)
        model.fit(X_train, EYE_gen_label)
        yhat = model.predict(X_test)
        acc_fake_mlp = accuracy_score(y_true=EYE_svc_label, y_pred = yhat)
        print("Fake Accuracy MLP:", acc_fake_mlp)
        model = RandomForestClassifier(n_estimators=100, random_state=1)
        model.fit(X_train, EYE_gen_label)
        yhat = model.predict(X_test)
        acc_fake_rf = accuracy_score(y_true=EYE_svc_label, y_pred = yhat)
        print("Fake Accuracy RF:", acc_fake_rf)
        # real data
        model2 = SVC(C=2, kernel = 'rbf', degree=3)
        X_train2 = np.concatenate([EEG_gen_test, EYE_gen_test], axis=1)
        model2.fit(X_train2, EYE_gen_label)
        yhat2 = model2.predict(X_test)
        acc_real_svc = accuracy_score(y_true=EYE_svc_label, y_pred = yhat2)
        print("Real Accuracy SVC:", acc_real_svc)
        model2 = MLPClassifier( activation='relu',max_iter=1000, random_state=1, alpha = 0.001, batch_size= 10)
        model2.fit(X_train2, EYE_gen_label)
        yhat2 = model2.predict(X_test)
        acc_real_mlp = accuracy_score(y_true=EYE_svc_label, y_pred = yhat2)
        print("Real Accuracy MLP:", acc_real_mlp)
        model2 = RandomForestClassifier(n_estimators=100, random_state=1)
        model2.fit(X_train2, EYE_gen_label)
        yhat2 = model2.predict(X_test)
        acc_real_rf = accuracy_score(y_true=EYE_svc_label, y_pred = yhat2)
        print("Real Accuracy RF:", acc_real_rf)
        # only eye data
        X_train3 = EYE_gen_test
        X_test3 = EYE_svc_test
        model3 = SVC(C=2, kernel = 'rbf', degree=3)
        model3.fit(X_train3, EYE_gen_label)
        yhat3 = model3.predict(X_test3)
        acc_eye_svc = accuracy_score(y_true=EYE_svc_label, y_pred = yhat3)
        print("Eye Accuracy SVC:", acc_eye_svc)
        model3 = MLPClassifier( activation='relu',max_iter=1000, random_state=1, alpha = 0.001, batch_size= 10)
        model3.fit(X_train3, EYE_gen_label)
        yhat3 = model3.predict(X_test3)
        acc_eye_mlp = accuracy_score(y_true=EYE_svc_label, y_pred = yhat3)
        print("Eye Accuracy MLP:", acc_eye_mlp)
        model3 = RandomForestClassifier(n_estimators=100, random_state=1)
        model3.fit(X_train3, EYE_gen_label)
        yhat3 = model3.predict(X_test3)
        acc_eye_rf = accuracy_score(y_true=EYE_svc_label, y_pred = yhat3)
        print("Eye Accuracy RF:", acc_eye_rf)
        log_string = str(acc_fake_svc)[:6]+","+str(acc_fake_mlp)[:6]+","+str(acc_fake_rf)[:6]+","+str(acc_real_svc)[:6]+","+str(acc_real_mlp)[:6]+","+str(acc_real_rf)[:6]+","+str(acc_eye_svc)[:6]+","+str(acc_eye_mlp)[:6]+","+str(acc_eye_rf)[:6]
        log_string += ","+transfer_method+","+str(idx)+","+str(session)
        logToFile(log_string)

if __name__ == "__main__":
    for idx in range(1,NUMBER_OF_SUBJECTS+1):
        for session in range(1,NUMBER_OF_SESSIONS+1):
            print("subject-based-cgan ",idx, "-",session,":")
            train("subject-based-cgan",idx,session)

