"""
Discriminator and Generator implementation for CWGAN-GP
"""

import torch
import torch.nn as nn

class Discriminator(nn.Module):
    def __init__(self, input_dim, discriminator_dim):
        super(Discriminator, self).__init__()
        seq = []
        for item in list(discriminator_dim):
            seq += [nn.Linear(input_dim, item), nn.LeakyReLU(0.2), nn.Dropout(0.5)]
            input_dim = item
        seq += [nn.Linear(input_dim, 1)]
        self.disc = nn.Sequential(*seq)

    def forward(self, x):
        return self.disc(x)

class DiscriminatorVanillia(nn.Module):
    def __init__(self, input_dim, discriminator_dim):
        super(DiscriminatorVanillia, self).__init__()
        seq = []
        for item in list(discriminator_dim):
            seq += [nn.Linear(input_dim, item), nn.LeakyReLU(0.2), nn.Dropout(0.5)]
            input_dim = item
        seq += [nn.Linear(input_dim, 1)]
        seq += [nn.Sigmoid()]
        self.disc = nn.Sequential(*seq)

    def forward(self, x):
        return self.disc(x)

class Residual(nn.Module):
    def __init__(self, i, o):
        super(Residual, self).__init__()
        self.fc = nn.Linear(i, o)
        self.bn = nn.BatchNorm1d(o)
        self.relu = nn.ReLU()

    def forward(self, input_):
        out = self.fc(input_)
        out = self.bn(out)
        out = self.relu(out)
        return torch.cat([out, input_], dim=1)

class Generator(nn.Module):
    def __init__(self, input_dim, generator_dim, output_dim):
        super(Generator, self).__init__()
        seq = []
        for item in list(generator_dim):
             seq += [Residual(input_dim, item)]
             input_dim += item
        seq.append(nn.Linear(input_dim, output_dim))
        self.net = nn.Sequential(*seq)

    def forward(self, x):
        return self.net(x)


def initialize_weights(model):
    # Initializes weights according to the DCGAN paper
    for m in model.modules():
        if isinstance(m, (nn.Conv2d, nn.ConvTranspose2d, nn.BatchNorm2d, nn.InstanceNorm2d)):
            nn.init.normal_(m.weight.data, 0.0, 0.02)

def weights_init(model):
    class_name = model.__class__.__name__
    if class_name.find('Conv') != -1:
        nn.init.normal_(model.weight.data, 0.0, 0.02)
    elif class_name.find('Linear') != -1:
        nn.init.normal_(model.weight.data, 0.0, 0.02)
        nn.init.normal_(model.bias.data, 0.0, 0.02)
    elif class_name.find('BatchNorm') != -1:
        nn.init.normal_(model.weight.data, 1.0, 0.02)
        nn.init.constant_(model.bias.data, 0.0)

def test():
    gen_input, gen_output, gen_dim, disc_input, disc_dim = 34, 310, (256,256), 311, (256,256)
    z = torch.randn((100, 34))
    x = torch.randn((100, 311))
    disc = Discriminator(disc_input, disc_dim)
    print(disc)
    output_disc = disc(x)
    print("output_disc.shape:",output_disc.shape)
    gen = Generator(gen_input, gen_dim, gen_output)
    print(gen)
    output_gen = gen(z)
    print("output_gen.shape:",output_gen.shape)


#test()